﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

using Infragistics.Shared;
using Infragistics.Win;
using Infragistics.Win.UltraWinGanttView;
using Infragistics.Win.UltraWinGrid;
using Infragistics.Win.UltraWinSchedule;
using Infragistics.Win.UltraWinSchedule.TaskUI;


namespace GanttSample
{
    public partial class Form1 : Form
    {
        private DataSet _resultSet;
        private DataSet _ganttData;
        private UltraCalendarInfo _calendarInfo;
        private List<DateInterval> _dateIntervalList;
        private int _currentDateIntervalIdx;

        private DateTime _oldTaskStartDate;
        private OriginalTaskData _originalTaskData;

        private Task _firstVisibleTask;
        private Task _lastActiveTask;
        private int _lastActiveTaskScrollOffset;

        private CreationFilter _creationFilter;
        private DrawFilter _drawFilter;

        private bool _firstTime;


        public Form1()
        {
            _firstVisibleTask = null;
            _lastActiveTask = null;
            _lastActiveTaskScrollOffset = -1;

            _creationFilter = new CreationFilter();
            _drawFilter = new DrawFilter();

            InitializeComponent();

            ultraGanttView.CreationFilter = _creationFilter;
            ultraGanttView.DrawFilter = _drawFilter;

            _firstTime = true;

            // create all allowed timeline intervals and set default (1 day)
            _dateIntervalList = CreateDateIntervalList();
            _currentDateIntervalIdx = 0;

            uneParentTasks.Value = 50;
            uneTasks.Value = 10;

            // customize the Gantt view
            CustomizeResources();
            CustomizeUI();
        }


        protected override void OnLoad(EventArgs e)
        {
            //_presenter.OnViewReady();
            base.OnLoad(e);

            //_originalTaskData = null;
        }


        #region Data binding
        public void Bind(DataSet data)
        {
            Stopwatch sw = new Stopwatch();
            sw.Start();

            _ganttData = data;
            _ganttData.Tables["Tasks"].DefaultView.Sort = "SortString ASC";

            if (_firstTime)
            {
                CreateDataBindings();
                _firstTime = false;
            }

            Debug.WriteLine("Bind: {0}ms", sw.ElapsedMilliseconds);
            sw.Stop();
        }


        private void CreateDataBindings()
        {
            Stopwatch sw = new Stopwatch();
            sw.Start();

            CreateDataBindingsForProjects();
            CreateDataBindingsForTasks();

            Debug.WriteLine(string.Format("DataBindings created: {0}ms", sw.ElapsedMilliseconds));
            sw.Restart();

            if (_calendarInfo.Projects.Count > 1)
                ultraGanttView.Project = _calendarInfo.Projects[1];

            Debug.WriteLine(string.Format("Project assigned: {0}ms", sw.ElapsedMilliseconds));
            sw.Restart();
        }


        private void CreateDataBindingsForTasks()
        {
            _calendarInfo.DataBindingsForTasks.BindingContextControl = this;

            _calendarInfo.DataBindingsForTasks.SetDataBinding(_ganttData, "Tasks");

            _calendarInfo.DataBindingsForTasks.IdMember = "Id";
            _calendarInfo.DataBindingsForTasks.NameMember = "Name";
            _calendarInfo.DataBindingsForTasks.StartDateTimeMember = "StartDate";
            _calendarInfo.DataBindingsForTasks.DurationMember = "Duration";
            _calendarInfo.DataBindingsForTasks.ProjectKeyMember = "ProjectKey";
            _calendarInfo.DataBindingsForTasks.ParentTaskIdMember = "ParentTaskId";
            _calendarInfo.DataBindingsForTasks.ConstraintMember = "Constraint";
            _calendarInfo.DataBindingsForTasks.ConstraintDateTimeMember = "ConstraintDateTime";
            _calendarInfo.DataBindingsForTasks.PercentCompleteMember = "PercentComplete";
            _calendarInfo.DataBindingsForTasks.NotesMember = "SortString";
        }


        private void CreateDataBindingsForProjects()
        {
            _calendarInfo.DataBindingsForProjects.BindingContextControl = this;

            _calendarInfo.DataBindingsForProjects.SetDataBinding(_ganttData, "Projects");

            _calendarInfo.DataBindingsForProjects.IdMember = "Id";
            _calendarInfo.DataBindingsForProjects.KeyMember = "ProjectKey";
            _calendarInfo.DataBindingsForProjects.NameMember = "Name";
            _calendarInfo.DataBindingsForProjects.StartDateMember = "StartDate";
        }
        #endregion


        #region UltraGanttView events
        void GanttTableView_Load(object sender, System.EventArgs e)
        {
            ultraGanttView.PropertyChanged += new PropertyChangedEventHandler(ultraGanttView_PropertyChanged);
            _calendarInfo.Tasks.SortedFields.Add(SortableTaskField.Notes, SortOrder.Ascending);
        }


        void ultraGanttView_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("ultraGanttView_PropertyChanged");
        }


        #region Debug
        int items = -1;
        public int GetItems() { return items; }
        long msecs = -1;
        public long GetMsecs() { return msecs; }
        public int GetItemCount()
        {
            int c = 0;
            foreach (Task t in ultraGanttView.CalendarInfo.Tasks)
            {
                c++;
                if (t.Tasks.Count > 0)
                    c += GetItemSubCount(t.Tasks);
            }

            return c;
        }
        private int GetItemSubCount(TasksCollection tCol)
        {
            int c = 0;
            foreach (Task t in tCol)
            {
                c++;
                if (t.Tasks.Count > 0)
                    c += GetItemSubCount(t.Tasks);
            }

            return c;
        }
        #endregion
        void ultraGanttView_InitializeTaskGridRow(object sender, InitializeTaskGridRowEventArgs e)
        {
            Stopwatch sw = new Stopwatch();
            sw.Start();

            try
            {
                DataRow dataRow = _ganttData.Tables["Tasks"].Rows.Find(e.Task.Id);
                if (dataRow != null)
                {
                    Guid taskDefId = Guid.Empty;
                    Guid.TryParse(dataRow["TaskDefinitionId"].ToString(), out taskDefId);

                    Guid responsibleUserId = Guid.Empty;
                    Guid.TryParse(dataRow["ResponsibleUserId"].ToString(), out responsibleUserId);

                    e.Task.SetCustomProperty(Identifier.COLNAME_TASKKEY, dataRow["Key"].ToString());
                    e.Task.SetCustomProperty(Identifier.COLNAME_RESPONSIBLEUSERID, responsibleUserId);
                    e.Task.SetCustomProperty(Identifier.COLNAME_RESPONSIBLEUSER, dataRow["ResponsibleUser"].ToString());
                    e.Task.SetCustomProperty(Identifier.COLNAME_TASKSTATE, dataRow["TaskExecutionState"].ToString());
                    e.Task.SetCustomProperty(Identifier.COLNAME_PATH, dataRow["Path"].ToString());
                    e.Task.SetCustomProperty(Identifier.COLNAME_TASKDEFID, taskDefId);
                }

                // reset start time to 00:00h to ensure task duration is exactly one day
                if (!e.Task.IsSummary)
                    e.Task.StartDateTime = e.Task.StartDateTime.Date;
            }
            catch (Exception exc)
            {
                //Logger.LogDebug("Unexpected error!", "UltraGanttView: An unexpected error occurred!\n" + exc.Message);
                System.Diagnostics.Debug.WriteLine("UltraGanttView: An unexpected error occurred! Task: " + e.Task.Name +
                    " | " + e.Task.StartDateTime + "\n" + exc.Message, "Unexpected error!");
            }

            Debug.WriteLine(string.Format("ultraGanttView_InitializeTaskGridRow: RNr: {4} @ {0}ms (Name: {1}, Key: {2}, Id: {3})",
                sw.ElapsedMilliseconds, e.Task.Name, "", e.Task.Id, e.Task.RowNumber));
            items++;
            msecs += sw.ElapsedMilliseconds;
        }


        private void ultraGanttView_TaskElementDragging(object sender, TaskElementDraggingEventArgs e)
        {
            if (e.Phase == TaskDragPhase.Beginning)
            {
                _originalTaskData = OriginalTaskData.BackupTaskData(e.Task);

                //if (RefreshTimerOff != null)
                //    RefreshTimerOff(this, new EventArgs<string>(Constants.UIExtensionSiteNames.GanttViewTab));
            }
        }


        private void ultraGanttView_TaskElementDragComplete(object sender, TaskElementDragCompleteEventArgs e)
        {
            Debug.WriteLine(string.Format("Task dropped @ {0}", DateTime.Now.ToLongTimeString()));

            if (HandleTaskAfterDragDrop(e.Task))
            {
                // ensure that tasks always start at 00:00h and occupy exactly one day
                if (e.Task.StartDateTime.TimeOfDay >= TimeSpan.FromHours(12))
                    e.Task.StartDateTime = e.Task.StartDateTime.AddDays(1).Date;
                else
                    e.Task.StartDateTime = e.Task.StartDateTime.Date;

                //_presenter.UpdateTask(e.Task);
            }

            //if (RefreshTimerOn != null)
            //    RefreshTimerOn(this, new EventArgs<string>(Constants.UIExtensionSiteNames.GanttViewTab));
        }


        private void ultraGanttView_CellActivating(object sender, CellActivatingEventArgs e)
        {
            e.Cancel = true;
        }


        private void ultraGanttView_ActiveTaskChanged(object sender, ActiveTaskChangedEventArgs e)
        {
            //if ((e.ActiveTask != null) && (!e.ActiveTask.IsSummary))
            //    _presenter.TaskExecutionSelected(e.ActiveTask.Id);
            //else
            //    _presenter.TaskExecutionSelected(null);
        }


        private void ultraGanttView_TaskDialogDisplaying(object sender, TaskDialogDisplayingEventArgs e)
        {
            //using (new WaitCursor())
            //{
                e.Cancel = true;

            //    if ((e.Task != null) && (!e.Task.IsSummary))
            //        _presenter.OpenTaskExecution(e.Task.Id);
            //}
        }
        #endregion


        #region Private methods
        private void CustomizeResources()
        {
            ResourceCustomizer rc = Infragistics.Win.UltraWinGanttView.Resources.Customizer;
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.RowNumber", Properties.Resources.GanttView_ColName_RowNumber);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Name", Properties.Resources.GanttView_ColName_Name);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.PercentComplete", Properties.Resources.GanttView_ColName_PercentComplete);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Constraint", Properties.Resources.GanttView_ColName_Constraint);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.ConstraintDateTime", Properties.Resources.GanttView_ColName_ConstraintDateTime);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Duration", Properties.Resources.GanttView_ColName_Duration);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.StartDateTime", Properties.Resources.GanttView_ColName_StartDateTime);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.EndDateTime", Properties.Resources.GanttView_ColName_EndDateTime);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Deadline", Properties.Resources.GanttView_ColName_Deadline);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Milestone", Properties.Resources.GanttView_ColName_Milestone);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Dependencies", Properties.Resources.GanttView_ColName_Dependencies);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Resources", Properties.Resources.GanttView_ColName_Resources);
            rc.SetCustomizedString("TaskProxy_PropertyDisplayName_Task.Notes", Properties.Resources.GanttView_ColName_Notes);
        }


        private void CustomizeUI()
        {
            _calendarInfo = new UltraCalendarInfo();

            #region Add custom columns
            _calendarInfo.CustomTaskColumns.Add(Identifier.COLNAME_TASKKEY, typeof(string), false, string.Empty);
            _calendarInfo.CustomTaskColumns.Add(Identifier.COLNAME_RESPONSIBLEUSERID, typeof(Guid), false, Guid.Empty);
            _calendarInfo.CustomTaskColumns.Add(Identifier.COLNAME_RESPONSIBLEUSER, typeof(string), false, string.Empty);
            _calendarInfo.CustomTaskColumns.Add(Identifier.COLNAME_TASKSTATE, typeof(int), false, 1);
            _calendarInfo.CustomTaskColumns.Add(Identifier.COLNAME_PATH, typeof(string), false, string.Empty);
            _calendarInfo.CustomTaskColumns.Add(Identifier.COLNAME_TASKDEFID, typeof(Guid), false, Guid.Empty);
            #endregion

            ultraGanttView.CalendarInfo = _calendarInfo;

            TaskColumnSettingsCollection colSettings = ultraGanttView.GridSettings.ColumnSettings;

            #region Set column visibility
            colSettings[TaskField.RowNumber].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Name].Visible = DefaultableBoolean.True;
            colSettings[TaskField.PercentComplete].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Constraint].Visible = DefaultableBoolean.False;
            colSettings[TaskField.ConstraintDateTime].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Duration].Visible = DefaultableBoolean.False;
            colSettings[TaskField.StartDateTime].Visible = DefaultableBoolean.True;
            colSettings[TaskField.EndDateTime].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Deadline].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Milestone].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Dependencies].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Resources].Visible = DefaultableBoolean.False;
            colSettings[TaskField.Notes].Visible = DefaultableBoolean.True;
            colSettings[Identifier.COLNAME_TASKKEY].Visible = DefaultableBoolean.True;
            colSettings[Identifier.COLNAME_RESPONSIBLEUSERID].Visible = DefaultableBoolean.False;
            colSettings[Identifier.COLNAME_RESPONSIBLEUSER].Visible = DefaultableBoolean.False;
            colSettings[Identifier.COLNAME_TASKSTATE].Visible = DefaultableBoolean.False;
            colSettings[Identifier.COLNAME_PATH].Visible = DefaultableBoolean.False;
            colSettings[Identifier.COLNAME_TASKDEFID].Visible = DefaultableBoolean.False;
            #endregion

            #region Set column position
            colSettings[TaskField.RowNumber].VisiblePosition = 0;
            colSettings[TaskField.Name].VisiblePosition = 1;
            colSettings[Identifier.COLNAME_TASKKEY].VisiblePosition = 2;
            colSettings[Identifier.COLNAME_PATH].VisiblePosition = 3;
            colSettings[TaskField.PercentComplete].VisiblePosition = 4;
            colSettings[TaskField.Constraint].VisiblePosition = 5;
            colSettings[TaskField.ConstraintDateTime].VisiblePosition = 6;
            colSettings[TaskField.Duration].VisiblePosition = 7;
            colSettings[TaskField.StartDateTime].VisiblePosition = 8;
            colSettings[TaskField.EndDateTime].VisiblePosition = 9;
            colSettings[TaskField.Deadline].VisiblePosition = 10;
            colSettings[TaskField.Milestone].VisiblePosition = 11;
            colSettings[TaskField.Dependencies].VisiblePosition = 12;
            colSettings[TaskField.Resources].VisiblePosition = 13;
            colSettings[TaskField.Notes].VisiblePosition = 14;
            #endregion

            #region Set column width
            colSettings[TaskField.RowNumber].Width = 0;
            colSettings[Identifier.COLNAME_TASKKEY].Width = 120;
            colSettings[TaskField.Name].Width = 350;
            colSettings[TaskField.StartDateTime].Width = 70;

            ultraGanttView.GridAreaWidth =
                colSettings[TaskField.RowNumber].Width +
                colSettings[Identifier.COLNAME_TASKKEY].Width +
                colSettings[TaskField.Name].Width +
                colSettings[TaskField.StartDateTime].Width;
            #endregion

            #region Timeline settings
            ultraGanttView.TimelineSettings.PrimaryInterval = _dateIntervalList[_currentDateIntervalIdx];

            ultraGanttView.TimelineSettings.BarSettings.BarHeight = 25;
            ultraGanttView.TimelineSettings.BarSettings.BarTextRight = BarTextField.ResourceNames;
            ultraGanttView.TimelineSettings.BarSettingsSummary.BarVisible = DefaultableBoolean.False;
            ultraGanttView.TimelineSettings.BarSettingsSummary.StartIndicatorVisible = DefaultableBoolean.False;
            ultraGanttView.TimelineSettings.BarSettingsSummary.EndIndicatorVisible = DefaultableBoolean.False;
            #endregion

            #region Grid settings
            ultraGanttView.GridSettings.RowSelectorHeaderStyle = RowSelectorHeaderStyle.None;
            #endregion

            #region Gantt table events
            //ultraGanttView.MouseWheel += new MouseEventHandler(ultraGanttView_MouseWheel);
            //ultraGanttView.TaskToolTipDisplaying += new TaskToolTipDisplayingHandler(ultraGanttView_TaskToolTipDisplaying);
            ultraGanttView.PropertyChanged += new PropertyChangedEventHandler(ultraGanttView_PropertyChanged);
            ultraGanttView.InitializeTaskGridRow += new InitializeTaskGridRowHandler(ultraGanttView_InitializeTaskGridRow);
            ultraGanttView.TaskElementDragging += new TaskElementDraggingHandler(ultraGanttView_TaskElementDragging);
            ultraGanttView.TaskElementDragComplete += new TaskElementDragCompleteHandler(ultraGanttView_TaskElementDragComplete);
            ultraGanttView.CellActivating += new CellActivatingHandler(ultraGanttView_CellActivating);
            ultraGanttView.ActiveTaskChanged += new ActiveTaskChangedHandler(ultraGanttView_ActiveTaskChanged);
            ultraGanttView.TaskDialogDisplaying += new TaskDialogDisplayingHandler(ultraGanttView_TaskDialogDisplaying);
            _calendarInfo.TaskConstraintViolation += new TaskConstraintViolationHandler(_calendarInfo_TaskConstraintViolation);
            #endregion
        }


        private void _calendarInfo_TaskConstraintViolation(object sender, TaskConstraintViolationEventArgs e)
        {
            e.PromptUser = false;
            e.Action = TaskConstraintViolationAction.Override;
        }


        private List<DateInterval> CreateDateIntervalList()
        {
            List<DateInterval> diList = new List<DateInterval>();

            diList.Add(new DateInterval(1, DateIntervalUnits.Days));
            diList.Add(new DateInterval(1, DateIntervalUnits.Weeks));
            diList.Add(new DateInterval(1, DateIntervalUnits.Months));

            return diList;
        }


        private bool HandleTaskAfterDragDrop(Task task)
        {
            if (_originalTaskData.TaskExecutionState < TaskExecutionState.Scheduled)
            {
                DialogResult dlgResult = MessageBox.Show("Nur eingeplante Aufträge können verschoben werden.\n" +
                    "Möchten Sie den Auftrag jetzt einplanen?", "Auftragsplanung", MessageBoxButtons.YesNo);
                if (dlgResult == DialogResult.Yes)
                    task.SetCustomProperty(Identifier.COLNAME_TASKSTATE, (int)TaskExecutionState.Scheduled);
                else
                    _originalTaskData.RestoreTaskData(ref task);

                return dlgResult == DialogResult.Yes;
            }

            return true;
        }


        private Task FindTask(TasksCollection taskCollection, Task task)
        {
            if (taskCollection == null || task == null || taskCollection.Count == 0)
                return null;

            Task parentTask = null;
            if (task.Parent != null)
            {
                parentTask = taskCollection.FirstOrDefault<Task>(t => t.Id.Equals(task.Parent.Id)) as Task;
                if (parentTask != null)
                    taskCollection = parentTask.Tasks;
            }

            return taskCollection.FirstOrDefault<Task>(t => t.Id.Equals(task.Id)) as Task;
        }
        #endregion


        public void BeforeDataUpdate()
        {
            // store last active task
            if (ultraGanttView.ActiveTask != null)
                _lastActiveTask = ultraGanttView.ActiveTask;
            else
                _lastActiveTask = null;

            // store first visible task
            RowColRegionIntersectionUIElement rowCol = ultraGanttView.UIElement.GetDescendant(
                typeof(RowColRegionIntersectionUIElement)) as RowColRegionIntersectionUIElement;
            UIElement row = rowCol.ChildElements[0];

            Task task = ultraGanttView.TaskFromPoint(new Point(row.Rect.X + 1, row.Rect.Y + 1));
            if (task != null)
                _firstVisibleTask = task;
            else
                _firstVisibleTask = null;

            // store active task scroll offset
            _lastActiveTaskScrollOffset = -1;
            if (_lastActiveTask != null && _firstVisibleTask != null)
                _lastActiveTaskScrollOffset = _lastActiveTask.RowNumber - _firstVisibleTask.RowNumber;

            ultraGanttView.BeginUpdate();
            ultraGanttView.SuspendLayout();
        }


        public void AfterDataUpdate()
        {
            Task foundTask = null;

            Stopwatch sw = new Stopwatch();
            sw.Start();

            // re-select formerly selected task
            if (_lastActiveTask != null)
            {
                foundTask = FindTask(_calendarInfo.Tasks, _lastActiveTask);
                if (foundTask != null)
                    ultraGanttView.ActiveTask = foundTask;
            }

            Debug.WriteLine(string.Format("AfterDataUpdate - ActiveTask: {0}ms", sw.ElapsedMilliseconds));
            sw.Restart();

            // re-set scrollbar position
            if (_firstVisibleTask != null)
            {
                foundTask = FindTask(_calendarInfo.Tasks, _firstVisibleTask);
                if (foundTask != null)
                    ultraGanttView.EnsureTaskInView(foundTask, true);
            }

            Debug.WriteLine(string.Format("AfterDataUpdate - FirstVisibleTask: {0}ms", sw.ElapsedMilliseconds));
            sw.Restart();

            ultraGanttView.ResumeLayout();
            ultraGanttView.EndUpdate();

            Debug.WriteLine(string.Format("AfterDataUpdate - ResumeLayout: {0}ms", sw.ElapsedMilliseconds));
            sw.Restart();
        }
        


        #region Additional code for the sample
        private void btnBind_Click(object sender, EventArgs e)
        {
            Debug.WriteLine("btnBind_Click()");

            if (_resultSet == null)
                _resultSet = new DataSet();

            DataTable pjTable;
            DataTable taskTable;

            Stopwatch sw = new Stopwatch();
            sw.Start();

            pjTable = FillProject();
            taskTable = FillTasks();

            Debug.WriteLine(string.Format("Data created in {0}ms", sw.ElapsedMilliseconds));


            if (_resultSet.Tables.Count == 0)
            {
                _resultSet.Tables.Add(pjTable);
                _resultSet.Tables.Add(taskTable);
            }
            else
            {
                DataTable tblTaskNew = taskTable;
                DataTable tblTaskOld = _resultSet.Tables["Tasks"].Copy();
                tblTaskOld.AcceptChanges();

                lock (_resultSet)
                {
                    Debug.WriteLine("Determine changes...");
                    // determine rows to insert, update and delete
                    List<DataRow> insertRowList = (from r in tblTaskNew.AsEnumerable()
                                                   where (tblTaskOld.Rows.Find(r["Id"]) == null)
                                                   select r).ToList();

                    List<DataRow> updateRowList = (from r in tblTaskNew.AsEnumerable()
                                                   where (tblTaskOld.Rows.Find(r["Id"]) != null)
                                                   select r).ToList();

                    List<DataRow> deleteRowList = (from r in tblTaskOld.AsEnumerable()
                                                   where (tblTaskNew.Rows.Find(r["Id"]) == null)
                                                   select r).ToList();

                    // update model with insert-, update- and deleteRows
                    foreach (DataRow deleteRow in deleteRowList)
                        deleteRow.Delete();

                    foreach (DataRow insertRow in insertRowList)
                        tblTaskOld.ImportRow(insertRow);

                    DataRow originalRow;
                    foreach (DataRow updateRow in updateRowList)
                    {
                        originalRow = tblTaskOld.Rows.Find(updateRow["Id"]);
                        foreach (DataColumn col in originalRow.Table.Columns)
                        {
                            if (!originalRow[col.Ordinal].Equals(updateRow[col.Ordinal]))
                                originalRow[col.Ordinal] = updateRow[col.Ordinal];
                        }
                    }

                    // get changes
                    DataTable changeTable = tblTaskOld.GetChanges(DataRowState.Added | DataRowState.Deleted | DataRowState.Modified);

                    Debug.WriteLine(string.Format("Changes processed in {0}ms ({1} changes)", sw.ElapsedMilliseconds,
                        changeTable != null ? changeTable.Rows.Count : 0));
                    sw.Restart();

                    if (changeTable != null)
                    {
                        changeTable.DefaultView.Sort = "SortString ASC";

                        Debug.WriteLine("ResultSet has " + (_resultSet.HasChanges() ? "" : "no ") + "uncommitted changes!");

                        // merge changes into binding dataset
                        _resultSet.Merge(changeTable);
                    }
                    else
                        Debug.WriteLine("No changes, thus skipping data merge!");

                    Debug.WriteLine("Merge: {0}ms", sw.ElapsedMilliseconds);
                }
            }

            Bind(_resultSet);
        }


        private DataTable FillProject()
        {
            DataTable pjTable;
            DataRow newRow;

            pjTable = CreateProjectDataTable();

            newRow = pjTable.NewRow();

            newRow["Id"] = Guid.NewGuid();
            newRow["ProjectKey"] = "ProjectKey";
            newRow["Name"] = "ProjectKey";
            newRow["StartDate"] = DateTime.Now;

            pjTable.Rows.Add(newRow);

            return pjTable;
        }


        private DataTable FillTasks()
        {
            Guid parentTaskId;
            DataTable taskTable;
            DataRow newRow;
            
            taskTable = CreateTaskDataTable();

            for (int i = 0; i < (int)uneParentTasks.Value; i++)
            {
                newRow = taskTable.NewRow();

                Guid.TryParse(string.Format("100{0:00000}-0000-0000-0000-000000000000", i), out parentTaskId);

                newRow["Id"] = parentTaskId;
                newRow["Name"] = "ParentTask " + i;
                newRow["StartDate"] = DateTime.Now.Date;
                newRow["Duration"] = TimeSpan.FromDays(1);
                newRow["ProjectKey"] = "ProjectKey";
                newRow["ParentTaskId"] = null;
                newRow["PercentComplete"] = 0;
                newRow["Constraint"] = TaskConstraint.MustStartOn;
                newRow["ConstraintDateTime"] = DateTime.Now.Date;
                newRow["SortString"] = parentTaskId.ToString();

                newRow["Key"] = newRow["Name"];
                newRow["ResponsibleUser"] = "User " + i;
                newRow["ResponsibleUserId"] = parentTaskId; // Guid.NewGuid();
                newRow["TaskExecutionState"] = 0;
                newRow["Path"] = parentTaskId;
                newRow["TaskDefinitionId"] = parentTaskId; // Guid.NewGuid();

                taskTable.Rows.Add(newRow);

                for (int j = 0; j < (int)uneTasks.Value; j++)
                {
                    newRow = taskTable.NewRow();

                    Guid g;
                    Guid.TryParse(string.Format("200{0:00000}-{1:0000}-0000-0000-000000000000", i, j), out g);
                    newRow["Id"] = g;

                    newRow["Name"] = "Task " + i + "." + j;
                    newRow["StartDate"] = DateTime.Now.Date;
                    newRow["Duration"] = TimeSpan.FromDays(1);
                    newRow["ProjectKey"] = "ProjectKey";
                    newRow["ParentTaskId"] = parentTaskId;
                    newRow["PercentComplete"] = 0;
                    newRow["Constraint"] = TaskConstraint.MustStartOn;
                    newRow["ConstraintDateTime"] = DateTime.Now.Date;
                    newRow["SortString"] = parentTaskId + "/" + g;

                    newRow["Key"] = "";
                    newRow["ResponsibleUser"] = "User " + i + "." + j;
                    newRow["ResponsibleUserId"] = g; // Guid.NewGuid();
                    newRow["TaskExecutionState"] = TaskExecutionState.Proposed;
                    newRow["Path"] = parentTaskId + "/" + g;
                    newRow["TaskDefinitionId"] = g; // Guid.NewGuid();

                    taskTable.Rows.Add(newRow);
                }
            }

            return taskTable;
        }
        #endregion


        #region Code moved from presenter class
        private DataTable CreateProjectDataTable()
        {
            DataTable pjTable = new DataTable("Projects");

            pjTable.Columns.Add("Id");
            pjTable.Columns.Add("ProjectKey");
            pjTable.Columns.Add("Name");
            pjTable.Columns.Add("StartDate", typeof(DateTime));

            pjTable.Columns["Id"].Unique = true;
            pjTable.PrimaryKey = new DataColumn[] { pjTable.Columns["Id"] };

            return pjTable;
        }


        private DataTable CreateTaskDataTable()
        {
            DataTable taskTable = new DataTable("Tasks");

            taskTable.Columns.Add("Id");
            taskTable.Columns.Add("Name");
            taskTable.Columns.Add("StartDate", typeof(DateTime));
            taskTable.Columns.Add("Duration", typeof(TimeSpan));
            taskTable.Columns.Add("ProjectKey");
            taskTable.Columns.Add("ParentTaskId");
            taskTable.Columns.Add("Constraint", typeof(object));
            taskTable.Columns.Add("ConstraintDateTime", typeof(DateTime));
            taskTable.Columns.Add("PercentComplete", typeof(int));
            taskTable.Columns.Add("SortString", typeof(string));

            taskTable.Columns.Add("Key", typeof(string));
            taskTable.Columns.Add("ResponsibleUser", typeof(string));
            taskTable.Columns.Add("ResponsibleUserId");
            taskTable.Columns.Add("TaskExecutionState", typeof(int));
            taskTable.Columns.Add("Path", typeof(string));
            taskTable.Columns.Add("TaskDefinitionId");

            taskTable.Columns["Id"].Unique = true;
            taskTable.PrimaryKey = new DataColumn[] { taskTable.Columns["Id"] };

            return taskTable;
        }
        #endregion

        private void ultraGanttView_Paint(object sender, PaintEventArgs e)
        {
            //Debug.WriteLine("Pain(t)!!!");

            if (items > 0 && msecs > 0)
            {
                Debug.WriteLine("Initialize: {0} items in {1}ms.", items, msecs);
                items = 0;
                msecs = 0;
            }
        }
    }



    class Identifier
    {
        public const string COLNAME_TASKKEY = "TaskKey";
        public const string COLNAME_RESPONSIBLEUSERID = "ResponsibleUserId";
        public const string COLNAME_RESPONSIBLEUSER = "ResponsibleUser";
        public const string COLNAME_TASKSTATE = "TaskExecutionState";
        public const string COLNAME_PATH = "Path";
        public const string COLNAME_TASKDEFID = "TaskDefinitionId";
    }



    class OriginalTaskData : Task
    {
        private Guid _originalId;
        public Guid OriginalId
        {
            get { return _originalId; }
        }


        private Guid _taskDefId;
        public Guid TaskDefId
        {
            get { return _taskDefId; }
        }


        private TaskExecutionState _taskState;
        public TaskExecutionState TaskExecutionState
        {
            get { return _taskState; }
        }


        private Guid _responsibleUserId;
        public Guid ResponsibleUserId
        {
            get { return _responsibleUserId; }
        }


        private string _responsibleUserName;
        public string ResponsibleUserName
        {
            get { return _responsibleUserName; }
        }


        private OriginalTaskData()
            : base()
        {

        }


        public static OriginalTaskData BackupTaskData(Task task)
        {
            OriginalTaskData copyData = new OriginalTaskData();
            copyData._originalId = task.Id;
            copyData._taskDefId = new Guid(task.GetCustomProperty(Identifier.COLNAME_TASKDEFID).ToString());
            copyData._responsibleUserId = new Guid(task.GetCustomProperty(Identifier.COLNAME_RESPONSIBLEUSERID).ToString());
            copyData._responsibleUserName = task.GetCustomProperty(Identifier.COLNAME_RESPONSIBLEUSER).ToString();
            copyData._taskState = (TaskExecutionState)int.Parse(task.GetCustomProperty(Identifier.COLNAME_TASKSTATE).ToString());
            copyData.StartDateTime = task.StartDateTime;

            return copyData;
        }


        public void RestoreTaskData(ref Task task)
        {
            if (task.Id == OriginalId)
            {
                task.SetCustomProperty(Identifier.COLNAME_RESPONSIBLEUSERID, ResponsibleUserId);
                task.SetCustomProperty(Identifier.COLNAME_RESPONSIBLEUSER, ResponsibleUserName);
                task.SetCustomProperty(Identifier.COLNAME_TASKSTATE, TaskExecutionState);
                task.StartDateTime = StartDateTime;
            }
        }        
    }

    [Flags]
    public enum TaskExecutionState : int
    {
        Undefined = 1,
        Proposed = 4,//2,4,8,16,32,64,128,256,512
        Scheduled = 16,//1024,2048,4096,8192,16384,32768,65536,131072,262144,524288
        InProcess = 64,//1048576,2097152,4194304,8388608,16777216,33554432,67108864,134217728,268435456,536870912,1073741824
        Performed = 256,
        Cancelled = 1024

    }
}
